from dxfeed.core.pxd_include.DXTypes cimport *
from dxfeed.core.pxd_include.EventData cimport *


cdef class FuncWrapper:
	cdef dxf_event_listener_t func
	@staticmethod
	cdef FuncWrapper make_from_ptr(dxf_event_listener_t f)

cdef void trade_default_listener(int event_type, dxf_const_string_t symbol_name,
			            		 const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void quote_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void summary_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void profile_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void time_and_sale_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void candle_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void order_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void greeks_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void theo_price_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void underlying_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void series_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil

cdef void configuration_default_listener(int event_type, dxf_const_string_t symbol_name,
			                     const dxf_event_data_t* data, int data_count, void* user_data) nogil


from setuptools import setup, Extension, find_packages
from setuptools.dist import Distribution
from pathlib import Path
import platform

try:
    from Cython.Build import cythonize
except ImportError:
    use_cython = False
    ext = 'c'
    ext_pp = 'cpp'
else:
    use_cython = True
    ext = ext_pp = 'pyx'

# Get all dxfeed c api c files to be compiled into separate lib
source_files_directory = Path(__file__).resolve().parent.joinpath('dxfeed', 'dxfeed-c-api', 'src')
source_files_paths = [str(path) for path in source_files_directory.glob('*.c')]
libs = list()
if platform.system() == 'Windows':
    source_files_paths.remove(str(source_files_directory.joinpath('Linux.c')))
    libs = ['ws2_32']
else:
    source_files_paths.remove(str(source_files_directory.joinpath('Win32.c')))

# Separate dxfeed c api library
dxfeed_c = ('dxfeed_c', {'sources': source_files_paths,
                         'libraries': libs})

extensions = [Extension('dxfeed.core.utils.helpers', ['dxfeed/core/utils/helpers.' + ext]),
              Extension('dxfeed.core.listeners.listener', ['dxfeed/core/listeners/listener.' + ext]),
              Extension('dxfeed.core.DXFeedPy', ['dxfeed/core/DXFeedPy.' + ext_pp], libraries=libs)]

if use_cython:
    extensions = cythonize(extensions, language_level=3)


def build(setup_kwargs):
    setup_kwargs.update({
        'ext_modules': extensions,
        'zip_safe': False,
        'libraries': [dxfeed_c],
        'packages': find_packages(),
        'include_dirs': ['dxfeed/dxfeed-c-api/include',
                         'dxfeed/dxfeed-c-api/src'],
    })

def build_extensions():
    """
    Function for building extensions inplace for docs and tests
    :return:
    """
    build_params = {}
    build(build_params)
    dist = Distribution(attrs=build_params)
    build_clib_cmd = dist.get_command_obj('build_clib')
    build_clib_cmd.ensure_finalized()
    build_clib_cmd.run()
    build_ext_cmd = dist.get_command_obj('build_ext')
    build_ext_cmd.ensure_finalized()
    build_ext_cmd.inplace = 1
    build_ext_cmd.run()



# /*
#  * The contents of this file are subject to the Mozilla Public License Version
#  * 1.1 (the "License") you may not use this file except in compliance with
#  * the License. You may obtain a copy of the License at
#  * http://www.mozilla.org/MPL/
#  *
#  * Software distributed under the License is distributed on an "AS IS" basis,
#  * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
#  * for the specific language governing rights and limitations under the
#  * License.
#  *
#  * The Initial Developer of the Original Code is Devexperts LLC.
#  * Portions created by the Initial Developer are Copyright (C) 2010
#  * the Initial Developer. All Rights Reserved.
#  *
#  * Contributor(s):
#  *
#  */
#  
# /*
#  *	Here we have what's related to the records we receive from the server.
#  */
 
#ifndef RECORD_DATA_H_INCLUDED
#define RECORD_DATA_H_INCLUDED

#include "DXTypes.h"
from dxfeed.core.pxd_include.DXTypes cimport *
# /* -------------------------------------------------------------------------- */
# /*
#  *	Record type constants
#  */
# /* -------------------------------------------------------------------------- */
cdef extern from "RecordData.h":
    ctypedef enum dx_record_info_id_t:
        dx_rid_begin = 0,
        dx_rid_trade = dx_rid_begin,
        dx_rid_quote,
        dx_rid_summary,
        dx_rid_profile,
        dx_rid_market_maker,
        dx_rid_order,
        dx_rid_time_and_sale,
        dx_rid_candle,
        dx_rid_trade_eth,
        dx_rid_spread_order,
        dx_rid_greeks,
        dx_rid_theo_price,
        dx_rid_underlying,
        dx_rid_series,
        dx_rid_configuration,
    
        # /* add new values above this line */
    
        dx_rid_count,
        dx_rid_invalid


    ctypedef dxf_int_t dx_record_id_t

# /* Invalid or empty record id */
#     ctypedef static const dx_record_id_t DX_RECORD_ID_INVALID = -1
    cdef dx_record_id_t DX_RECORD_ID_INVALID = -1

# /* -------------------------------------------------------------------------- */
# /*
#  *	Record structures
#  */
# /* -------------------------------------------------------------------------- */

    ctypedef struct dx_trade_t:
        dxf_int_t time
        dxf_int_t sequence
        dxf_int_t time_nanos
        dxf_char_t exchange_code
        dxf_double_t price
        dxf_int_t size
        dxf_int_t tick
        dxf_double_t change
        dxf_int_t flags
        dxf_double_t day_volume
        dxf_double_t day_turnover

    ctypedef dx_trade_t dx_trade_eth_t

    ctypedef struct dx_quote_t:
        dxf_int_t sequence
        dxf_int_t time_nanos
        dxf_int_t bid_time
        dxf_char_t bid_exchange_code
        dxf_double_t bid_price
        dxf_int_t bid_size
        dxf_int_t ask_time
        dxf_char_t ask_exchange_code
        dxf_double_t ask_price
        dxf_int_t ask_size

    ctypedef struct dx_summary_t:
        dxf_dayid_t day_id
        dxf_double_t day_open_price
        dxf_double_t day_high_price
        dxf_double_t day_low_price
        dxf_double_t day_close_price
        dxf_dayid_t prev_day_id
        dxf_double_t prev_day_close_price
        dxf_double_t prev_day_volume
        dxf_int_t open_interest
        dxf_int_t flags


    ctypedef struct dx_profile_t:
        dxf_double_t beta
        dxf_double_t eps
        dxf_int_t div_freq
        dxf_double_t exd_div_amount
        dxf_dayid_t exd_div_date
        dxf_double_t _52_high_price
        dxf_double_t _52_low_price
        dxf_double_t shares
        dxf_double_t free_float
        dxf_double_t high_limit_price
        dxf_double_t low_limit_price
        dxf_int_t halt_start_time
        dxf_int_t halt_end_time
        dxf_int_t flags
        dxf_const_string_t description
        dxf_const_string_t status_reason

cdef extern from "RecordData.h":
    ctypedef struct dx_market_maker_t:
        dxf_char_t mm_exchange
        dxf_int_t mm_id
        dxf_int_t mmbid_time
        dxf_double_t mmbid_price
        dxf_int_t mmbid_size
        dxf_int_t mmbid_count
        dxf_int_t mmask_time
        dxf_double_t mmask_price
        dxf_int_t mmask_size
        dxf_int_t mmask_count

    ctypedef struct dx_order_t:
        dxf_int_t index
        dxf_int_t time
        dxf_int_t time_nanos
        dxf_int_t sequence
        dxf_double_t price
        dxf_int_t size
        dxf_int_t count
        dxf_int_t flags
        dxf_int_t mmid


    ctypedef struct dx_spread_order_t:
        dxf_int_t index
        dxf_int_t time
        dxf_int_t time_nanos
        dxf_int_t sequence
        dxf_double_t price
        dxf_int_t size
        dxf_int_t count
        dxf_int_t flags
        dxf_const_string_t spread_symbol


    ctypedef struct dx_time_and_sale_t:
        dxf_int_t time
        dxf_int_t sequence
        dxf_char_t exchange_code
        dxf_double_t price
        dxf_int_t size
        dxf_double_t bid_price
        dxf_double_t ask_price
        dxf_int_t exchange_sale_conditions
        dxf_int_t flags
        dxf_const_string_t buyer
        dxf_const_string_t seller

    ctypedef struct dx_candle_t:
        dxf_int_t time
        dxf_int_t sequence
        dxf_double_t count
        dxf_double_t open
        dxf_double_t high
        dxf_double_t low
        dxf_double_t close
        dxf_double_t volume
        dxf_double_t vwap
        dxf_double_t bid_volume
        dxf_double_t ask_volume
        dxf_int_t open_interest
        dxf_double_t imp_volatility

    ctypedef struct dx_greeks_t:
        dxf_int_t time
        dxf_int_t sequence
        dxf_double_t price
        dxf_double_t volatility
        dxf_double_t delta
        dxf_double_t gamma
        dxf_double_t theta
        dxf_double_t rho
        dxf_double_t vega

    ctypedef struct dx_theo_price_t:
        # // To have same record for record and event
        dxf_long_t time
        dxf_double_t price
        dxf_double_t underlying_price
        dxf_double_t delta
        dxf_double_t gamma
        dxf_double_t dividend
        dxf_double_t interest

    ctypedef struct dx_underlying_t:
        dxf_double_t volatility
        dxf_double_t front_volatility
        dxf_double_t back_volatility
        dxf_double_t put_call_ratio

    ctypedef struct dx_series_t:
        dxf_int_t index
        dxf_int_t time
        dxf_int_t sequence
        dxf_dayid_t expiration
        dxf_double_t volatility
        dxf_double_t put_call_ratio
        dxf_double_t forward_price
        dxf_double_t dividend
        dxf_double_t interest


    ctypedef struct dx_configuration_t:
        dxf_int_t version
        dxf_byte_array_t object

#endif /* RECORD_DATA_H_INCLUDED */


# something about license

#ifndef DX_TYPES_H_INCLUDED
#define DX_TYPES_H_INCLUDED

cdef extern from "DXTypes.h":
    ctypedef int ERRORCODE
    ctypedef void* dxf_subscription_t
    ctypedef void* dxf_connection_t
    ctypedef void* dxf_candle_attributes_t
    ctypedef void* dxf_snapshot_t
    ctypedef void* dxf_price_level_book_t
    ctypedef void* dxf_regional_book_t

#ifdef _WIN32

#pxd_include <wchar.h>
from libc.stddef cimport wchar_t

cdef extern from "DXTypes.h":
    ctypedef unsigned char      dxf_bool_t           # 8 bit
    ctypedef char               dxf_byte_t           # 8 bit
    ctypedef unsigned char      dxf_ubyte_t  # 8 bit
    ctypedef wchar_t            dxf_char_t           # 16 bit
    #ctypedef unsigned wchar_t   dx_unsigned_char_t  # 16 bit
    ctypedef short int          dxf_short_t          # 16 bit
    ctypedef unsigned short int dxf_ushort_t # 16 bit
    ctypedef int                dxf_int_t            # 32 bit
    ctypedef unsigned int       dxf_uint_t   # 32 bit
    ctypedef float              dxf_float_t          # 32 bit
    ctypedef long long          dxf_long_t           # 64 bit
    ctypedef unsigned long long dxf_ulong_t  # 64 bit
    ctypedef double             dxf_double_t         # 64 bit
    ctypedef int                dxf_dayid_t

    ctypedef dxf_char_t*        dxf_string_t
    ctypedef const dxf_char_t*  dxf_const_string_t

#else /* POSIX? */

#pxd_include <stdint.h>
#pxd_include <wchar.h>

# ctypedef unsigned char    dxf_bool_t           # 8 bit
# ctypedef int8_t           dxf_byte_t           # 8 bit
# ctypedef uint8_t          dxf_ubyte_t  # 8 bit
# ctypedef wchar_t          dxf_char_t           # 16 bit
# #ctypedef unsigned wchar_t   dx_unsigned_char_t  # 16 bit
# ctypedef int16_t          dxf_short_t          # 16 bit
# ctypedef uint16_t         dxf_ushort_t # 16 bit
# ctypedef int32_t          dxf_int_t            # 32 bit
# ctypedef uint32_t         dxf_uint_t   # 32 bit
# ctypedef float            dxf_float_t          # 32 bit
# ctypedef int64_t          dxf_long_t           # 64 bit
# ctypedef uint64_t         dxf_ulong_t  # 64 bit
# ctypedef double           dxf_double_t         # 64 bit
# ctypedef int32_t          dxf_dayid_t
#
# ctypedef dxf_char_t*        dxf_string_t
# ctypedef const dxf_char_t*  dxf_const_string_t

#endif /* _WIN32/POSIX */

cdef extern from "DXTypes.h":
    ctypedef dxf_uint_t dxf_event_flags_t

    ctypedef struct dxf_byte_array_t:
        dxf_byte_t* elements,
        int size,
        int capacity,

    ctypedef struct dxf_property_item_t:
        dxf_string_t key,
        dxf_string_t value,

    ctypedef enum dxf_connection_status_t:
        dxf_cs_not_connected = 0,
        dxf_cs_connected,
        dxf_cs_login_required,
        dxf_cs_authorized

#endif /* DX_TYPES_H_INCLUDED */


from warnings import warn
from dxfeed.core.pxd_include.DXErrorCodes cimport *
from dxfeed.core.pxd_include.DXFeed cimport *
from dxfeed.core.pxd_include.DXTypes cimport *
from pathlib import Path
import dxfeed

cdef object unicode_from_dxf_const_string_t(dxf_const_string_t wcs):
    """
    Cython function, callable from C to convert dxf_const_string_t(wchar_t) to unicode
    
    Parameters
    ----------
    wcs: dxf_const_string_t, wchar_t
        C string
        
    Returns
    -------
    unicode
        Python unicode string
    """
    if wcs == NULL:
      return ''
    ret_val = <object>PyUnicode_FromWideChar(wcs, -1)
    return ret_val

cdef dxf_const_string_t dxf_const_string_t_from_unicode(object symbol):
    """
    Cython function, callable from C that converts python unicode string to wchar_t. 
    
    Parameters
    ----------
    symbol: unicode
        Unicode string, usually ticker
    Returns
    -------
    wchar_t
        Wide Char
    """
    return PyUnicode_AsWideCharString(symbol, NULL)

def event_type_convert(event_type: str):
    """
    The function converts event type given as string to int, used in C dxfeed dxfeed

    Parameters
    ----------
    event_type: str
        Event type: 'Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH', 'SpreadOrder',
                    'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration' or ''
    Returns
    -------
    : int
        Integer that corresponds to event type in C dxfeed dxfeed
    """
    et_mapping = {
        'Trade': 1,
        'Quote': 2,
        'Summary': 4,
        'Profile': 8,
        'Order': 16,
        'TimeAndSale': 32,
        'Candle': 64,
        'TradeETH': 128,
        'SpreadOrder': 256,
        'Greeks': 512,
        'TheoPrice': 1024,
        'Underlying': 2048,
        'Series': 4096,
        'Configuration': 8192,
        '': -16384
    }
    try:
        return et_mapping[event_type]
    except KeyError:
        warn(f'No mapping for {event_type}, please choose one from {et_mapping.keys()}')
        return -16384

def get_include():
    """
    Function returns paths to header files of dxfeed-c-api library. Used for writing custom listeners

    Returns
    -------
    out_dir: list
        List of paths to header files
    """
    out_dir = list()
    out_dir.append(str(Path(dxfeed.__file__).resolve().parent.joinpath('dxfeed-c-api', 'include')))
    return out_dir


from dxfeed.core.utils.helpers cimport *

cdef class FuncWrapper:
    def __cinit__(self):
        self.func = NULL

    @staticmethod
    cdef FuncWrapper make_from_ptr(dxf_event_listener_t f):
        cdef FuncWrapper out = FuncWrapper()
        out.func = f
        return out

TRADE_COLUMNS = ['Symbol', 'Price', 'ExchangeCode', 'Size', 'Tick', 'Change', 'DayVolume', 'Time', 'IsETH']
cdef void trade_default_listener(int event_type,
                                 dxf_const_string_t symbol_name,
                                 const dxf_event_data_t*data,
                                 int data_count, void*user_data) nogil:
    cdef dxf_trade_t*trades = <dxf_trade_t*> data
    with gil:
        py_data = <object> user_data

        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 trades[i].price,
                                 trades[i].exchange_code,
                                 trades[i].size,
                                 trades[i].tick,
                                 trades[i].change,
                                 trades[i].day_volume,
                                 trades[i].time,
                                 trades[i].is_eth])

QUOTE_COLUMNS = ['Symbol', 'BidTime', 'BidExchangeCode', 'BidPrice', 'BidSize', 'AskTime', 'AskExchangeCode',
                 'AskPrice', 'AskSize', 'Scope']
cdef void quote_default_listener(int event_type,
                                 dxf_const_string_t symbol_name,
                                 const dxf_event_data_t*data,
                                 int data_count,
                                 void*user_data) nogil:
    cdef dxf_quote_t*quotes = <dxf_quote_t*> data
    with gil:
        py_data = <object> user_data

        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 quotes[i].bid_time,
                                 quotes[i].bid_exchange_code,
                                 quotes[i].bid_price,
                                 quotes[i].bid_size,
                                 quotes[i].ask_time,
                                 quotes[i].ask_exchange_code,
                                 quotes[i].ask_price,
                                 quotes[i].ask_size,
                                 <int> quotes[i].scope])

SUMMARY_COLUMNS = ['Symbol', 'DayId', 'DayHighPrice', 'DayLowPrice', 'DayClosePrice', 'PrevDayId', 'PrevDayClosePrice',
                   'PrevDayVolume', 'OpenInterest']
cdef void summary_default_listener(int event_type, dxf_const_string_t symbol_name,
                                   const dxf_event_data_t*data, int data_count, void*user_data) nogil:
    cdef dxf_summary_t*summary = <dxf_summary_t*> data

    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 summary[i].day_id,
                                 summary[i].day_high_price,
                                 summary[i].day_low_price,
                                 summary[i].day_close_price,
                                 summary[i].prev_day_id,
                                 summary[i].prev_day_close_price,
                                 summary[i].prev_day_volume,
                                 summary[i].open_interest])

PROFILE_COLUMNS = ['Symbol', 'Beta', 'EPS', 'DivFreq', 'ExdDivAmount', 'ExdDivDate', '52HighPrice', '52LowPrice',
                   'Shares', 'Description', 'RawFlags', 'StatusReason']
cdef void profile_default_listener(int event_type,
                                   dxf_const_string_t symbol_name,
                                   const dxf_event_data_t*data,
                                   int data_count,
                                   void*user_data)  nogil:
    cdef dxf_profile_t*p = <dxf_profile_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 p[i].beta,
                                 p[i].eps,
                                 p[i].div_freq,
                                 p[i].exd_div_amount,
                                 p[i].exd_div_date,
                                 p[i]._52_high_price,
                                 p[i]._52_low_price,
                                 p[i].shares,
                                 unicode_from_dxf_const_string_t(p[i].description),
                                 p[i].raw_flags,
                                 unicode_from_dxf_const_string_t(p[i].status_reason)])

TIME_AND_SALE_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'ExchangeCode', 'Price', 'Size', 'BidPrice',
                         'AskPrice', 'ExchangeSaleConditions', 'RawFlags', 'Buyer', 'Seller', 'Side', 'Type',
                         'IsValidTick', 'IsEthTrade', 'TradeThroughExempt', 'IsSpreadLeg']
cdef void time_and_sale_default_listener(int event_type,
                                         dxf_const_string_t symbol_name,
                                         const dxf_event_data_t*data,
                                         int data_count,
                                         void*user_data) nogil:
    cdef dxf_time_and_sale_t*tns = <dxf_time_and_sale_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 tns[i].event_flags,
                                 tns[i].index,
                                 tns[i].time,
                                 tns[i].exchange_code,
                                 tns[i].price,
                                 tns[i].size,
                                 tns[i].bid_price,
                                 tns[i].ask_price,
                                 unicode_from_dxf_const_string_t(tns[i].exchange_sale_conditions),
                                 tns[i].raw_flags,
                                 unicode_from_dxf_const_string_t(tns[i].buyer),
                                 unicode_from_dxf_const_string_t(tns[i].seller),
                                 tns[i].side,
                                 tns[i].type,
                                 tns[i].is_valid_tick,
                                 tns[i].is_eth_trade,
                                 tns[i].trade_through_exempt,
                                 tns[i].is_spread_leg])

CANDLE_COLUMNS = ['Symbol', 'Index', 'Time', 'Sequence', 'Count', 'Open', 'High', 'Low', 'Close', 'Volume', 'VWap',
                  'BidVolume', 'AskVolume', 'OpenInterest', 'ImpVolatility']
cdef void candle_default_listener(int event_type,
                                  dxf_const_string_t symbol_name,
                                  const dxf_event_data_t*data,
                                  int data_count,
                                  void*user_data) nogil:
    cdef dxf_candle_t*candle = <dxf_candle_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 candle[i].index,
                                 candle[i].time,
                                 candle[i].sequence,
                                 candle[i].count,
                                 candle[i].open,
                                 candle[i].high,
                                 candle[i].low,
                                 candle[i].close,
                                 candle[i].volume,
                                 candle[i].vwap,
                                 candle[i].bid_volume,
                                 candle[i].ask_volume,
                                 candle[i].open_interest,
                                 candle[i].imp_volatility])

ORDER_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'TimeNanos', 'Sequence', 'Price', 'Size', 'Count', 'Scope',
                 'Side', 'ExchangeCode', 'MarketMaker', 'SpreadSymbol']
cdef void order_default_listener(int event_type,
                                 dxf_const_string_t symbol_name,
                                 const dxf_event_data_t*data,
                                 int data_count,
                                 void*user_data) nogil:
    cdef dxf_order_t*order = <dxf_order_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 order[i].event_flags,
                                 order[i].index,
                                 order[i].time,
                                 order[i].time_nanos,
                                 order[i].sequence,
                                 order[i].price,
                                 order[i].size,
                                 order[i].count,
                                 order[i].scope,
                                 order[i].side,
                                 order[i].exchange_code,
                                 unicode_from_dxf_const_string_t(order[i].market_maker),
                                 unicode_from_dxf_const_string_t(order[i].spread_symbol)])

GREEKS_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Price', 'Volatility', 'Delta', 'Gamma', 'Theta', 'Rho',
                  'Vega']
cdef void greeks_default_listener(int event_type,
                                  dxf_const_string_t symbol_name,
                                  const dxf_event_data_t*data,
                                  int data_count,
                                  void*user_data) nogil:
    cdef dxf_greeks_t*greeks = <dxf_greeks_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 greeks[i].event_flags,
                                 greeks[i].index,
                                 greeks[i].time,
                                 greeks[i].price,
                                 greeks[i].volatility,
                                 greeks[i].delta,
                                 greeks[i].gamma,
                                 greeks[i].theta,
                                 greeks[i].rho,
                                 greeks[i].vega])

THEO_PRICE_COLUMNS = ['Symbol', 'Time', 'Price', 'UnderlyingPrice', 'Delta', 'Gamma', 'Dividend', 'Interest']
cdef void theo_price_default_listener(int event_type,
                                      dxf_const_string_t symbol_name,
                                      const dxf_event_data_t*data,
                                      int data_count,
                                      void*user_data) nogil:
    cdef dxf_theo_price_t*theo_price = <dxf_theo_price_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 theo_price[i].time,
                                 theo_price[i].price,
                                 theo_price[i].underlying_price,
                                 theo_price[i].delta,
                                 theo_price[i].gamma,
                                 theo_price[i].dividend,
                                 theo_price[i].interest])

UNDERLYING_COLUMNS = ['Symbol', 'Volatility', 'FrontVolatility', 'BackVolatility', 'PutCallRatio']
cdef void underlying_default_listener(int event_type,
                                      dxf_const_string_t symbol_name,
                                      const dxf_event_data_t*data,
                                      int data_count,
                                      void*user_data) nogil:
    cdef dxf_underlying_t*underlying = <dxf_underlying_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 underlying[i].volatility,
                                 underlying[i].front_volatility,
                                 underlying[i].back_volatility,
                                 underlying[i].put_call_ratio])

SERIES_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Sequence', 'Sequence', 'Expiration', 'Volatility',
                  'PutCallRatio', 'ForwardPrice', 'Dividend', 'Interest']
cdef void series_default_listener(int event_type,
                                  dxf_const_string_t symbol_name,
                                  const dxf_event_data_t*data,
                                  int data_count,
                                  void*user_data) nogil:
    cdef dxf_series_t*series = <dxf_series_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 series[i].event_flags,
                                 series[i].index,
                                 series[i].time,
                                 series[i].sequence,
                                 series[i].sequence,
                                 series[i].expiration,
                                 series[i].volatility,
                                 series[i].put_call_ratio,
                                 series[i].forward_price,
                                 series[i].dividend,
                                 series[i].interest])

CONFIGURATION_COLUMNS = ['Symbol', 'Version', 'Object']
cdef void configuration_default_listener(int event_type,
                                         dxf_const_string_t symbol_name,
                                         const dxf_event_data_t*data,
                                         int data_count,
                                         void*user_data) nogil:
    cdef dxf_configuration_t*config = <dxf_configuration_t*> data
    with gil:
        py_data = <object> user_data
        for i in range(data_count):
            py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
                                 config[i].version,
                                 unicode_from_dxf_const_string_t(config[i].object)])



from dxfeed.core.pxd_include.DXTypes cimport dxf_const_string_t
from cpython.ref cimport PyObject

cdef extern from *:
    PyObject * PyUnicode_FromWideChar(dxf_const_string_t w, Py_ssize_t size)
    dxf_const_string_t PyUnicode_AsWideCharString(object, Py_ssize_t *)

cdef object unicode_from_dxf_const_string_t(dxf_const_string_t wcs)

cdef dxf_const_string_t dxf_const_string_t_from_unicode(object symbol)


# something about license
 
#ifndef EVENT_DATA_H_INCLUDED
#define EVENT_DATA_H_INCLUDED

#pxd_include "RecordData.h"
from dxfeed.core.pxd_include.RecordData cimport *
#pxd_include "DXTypes.h"
from dxfeed.core.pxd_include.DXTypes cimport *
#ifndef OUT
    #define OUT
#endif /* OUT */

# /* -------------------------------------------------------------------------- */
# /*
#  *	Event type constants
#  */
# /* -------------------------------------------------------------------------- */
cdef extern from "EventData.h":
    enum dx_event_id_t:
        dx_eid_begin = 0,
        dx_eid_trade = dx_eid_begin,
        dx_eid_quote,
        dx_eid_summary,
        dx_eid_profile,
        dx_eid_order,
        dx_eid_time_and_sale,
        dx_eid_candle,
        dx_eid_trade_eth,
        dx_eid_spread_order,
        dx_eid_greeks,
        dx_eid_theo_price,
        dx_eid_underlying,
        dx_eid_series,
        dx_eid_configuration,

        # /* add new event id above this line */

        dx_eid_count,
        dx_eid_invalid


#define DXF_ET_TRADE         (1 << dx_eid_trade)
#define DXF_ET_QUOTE         (1 << dx_eid_quote)
#define DXF_ET_SUMMARY       (1 << dx_eid_summary)
#define DXF_ET_PROFILE       (1 << dx_eid_profile)
#define DXF_ET_ORDER         (1 << dx_eid_order)
#define DXF_ET_TIME_AND_SALE (1 << dx_eid_time_and_sale)
#define DXF_ET_CANDLE        (1 << dx_eid_candle)
#define DXF_ET_TRADE_ETH     (1 << dx_eid_trade_eth)
#define DXF_ET_SPREAD_ORDER  (1 << dx_eid_spread_order)
#define DXF_ET_GREEKS        (1 << dx_eid_greeks)
#define DXF_ET_THEO_PRICE    (1 << dx_eid_theo_price)
#define DXF_ET_UNDERLYING    (1 << dx_eid_underlying)
#define DXF_ET_SERIES        (1 << dx_eid_series)
#define DXF_ET_CONFIGURATION (1 << dx_eid_configuration)
#define DXF_ET_UNUSED        (~((1 << dx_eid_count) - 1))

#define DX_EVENT_BIT_MASK(event_id) (1 << event_id)

# // The length of record suffix including including the terminating null character
#define DXF_RECORD_SUFFIX_SIZE 5
DEF DXF_RECORD_SUFFIX_SIZE = 5
# /* -------------------------------------------------------------------------- */
# /*
# *	Source suffix array
# */
# /* -------------------------------------------------------------------------- */
cdef extern from "EventData.h":
    struct dx_suffix_t:
        dxf_char_t suffix[DXF_RECORD_SUFFIX_SIZE]


    struct dx_order_source_array_t:
        dx_suffix_t *elements,
        size_t size,
        size_t capacity


    ctypedef dx_order_source_array_t* dx_order_source_array_ptr_t

# /* -------------------------------------------------------------------------- */
# /*
#  *	Event data structures and support
#  */
# /* -------------------------------------------------------------------------- */

    ctypedef void* dxf_event_data_t

    ctypedef enum dxf_order_scope_t:
        dxf_osc_composite = 0,
        dxf_osc_regional = 1,
        dxf_osc_aggregate = 2,
        dxf_osc_order = 3


# /* Trade & Trade ETH -------------------------------------------------------- */

    ctypedef enum dxf_direction_t:
        dxf_dir_undefined = 0,
        dxf_dir_down = 1,
        dxf_dir_zero_down = 2,
        dxf_dir_zero = 3,
        dxf_dir_zero_up = 4,
        dxf_dir_up = 5


    ctypedef struct dxf_trade_t:
        dxf_long_t time,
        dxf_int_t sequence,
        dxf_int_t time_nanos,
        dxf_char_t exchange_code,
        dxf_double_t price,
        dxf_int_t size,
        # /* This field is absent in TradeETH */
        dxf_int_t tick,
        # /* This field is absent in TradeETH */
        dxf_double_t change,
        dxf_int_t raw_flags,
        dxf_double_t day_volume,
        dxf_double_t day_turnover,
        dxf_direction_t direction,
        dxf_bool_t is_eth,
        dxf_order_scope_t scope


# /* Quote -------------------------------------------------------------------- */

    ctypedef struct dxf_quote_t:
        dxf_long_t time,
        dxf_int_t sequence,
        dxf_int_t time_nanos,
        dxf_long_t bid_time,
        dxf_char_t bid_exchange_code,
        dxf_double_t bid_price,
        dxf_int_t bid_size,
        dxf_long_t ask_time,
        dxf_char_t ask_exchange_code,
        dxf_double_t ask_price,
        dxf_int_t ask_size,
        dxf_order_scope_t scope


# /* Summary ------------------------------------------------------------------ */

    ctypedef enum dxf_price_type_t:
        dxf_pt_regular = 0,
        dxf_pt_indicative = 1,
        dxf_pt_preliminary = 2,
        dxf_pt_final = 3


    ctypedef struct dxf_summary_t:
        dxf_dayid_t day_id,
        dxf_double_t day_open_price,
        dxf_double_t day_high_price,
        dxf_double_t day_low_price,
        dxf_double_t day_close_price,
        dxf_dayid_t prev_day_id,
        dxf_double_t prev_day_close_price,
        dxf_double_t prev_day_volume,
        dxf_int_t open_interest,
        dxf_int_t raw_flags,
        dxf_char_t exchange_code,
        dxf_price_type_t day_close_price_type,
        dxf_price_type_t prev_day_close_price_type


# /* Profile ------------------------------------------------------------------ */

    ctypedef enum dxf_trading_status_t:
      dxf_ts_undefined = 0,
      dxf_ts_halted = 1,
      dxf_ts_active = 2


    ctypedef enum dxf_short_sale_restriction_t:
      dxf_ssr_undefined = 0,
      dxf_ssr_active = 1,
      dxf_ssr_inactive = 2


    ctypedef struct dxf_profile_t:
        dxf_double_t beta,
        dxf_double_t eps,
        dxf_int_t div_freq,
        dxf_double_t exd_div_amount,
        dxf_dayid_t exd_div_date,
        dxf_double_t _52_high_price,
        dxf_double_t _52_low_price,
        dxf_double_t shares,
        dxf_double_t free_float,
        dxf_double_t high_limit_price,
        dxf_double_t low_limit_price,
        dxf_long_t halt_start_time,
        dxf_long_t halt_end_time,
        dxf_int_t raw_flags,
        dxf_const_string_t description,
        dxf_const_string_t status_reason,
        dxf_trading_status_t trading_status,
        dxf_short_sale_restriction_t ssr


# /* Order & Spread Order ----------------------------------------------------- */

    ctypedef enum dxf_order_side_t:
        dxf_osd_undefined = 0,
        dxf_osd_buy = 1,
        dxf_osd_sell = 2

    # union _inner_oso:
    #     dxf_const_string_t market_maker,
    #     dxf_const_string_t spread_symbol


    ctypedef struct dxf_order_t:
        dxf_event_flags_t event_flags,
        dxf_long_t index,
        dxf_long_t time,
        dxf_int_t time_nanos,
        dxf_int_t sequence,
        dxf_double_t price,
        dxf_int_t size,
        dxf_int_t count,
        dxf_order_scope_t scope,
        dxf_order_side_t side,
        dxf_char_t exchange_code,
        dxf_char_t source[DXF_RECORD_SUFFIX_SIZE],
        # _inner_oso # probably there should be some name (p58 of cython book)
        dxf_const_string_t market_maker,
        dxf_const_string_t spread_symbol

# /* Time And Sale ------------------------------------------------------------ */

    ctypedef enum dxf_tns_type_t:
        dxf_tnst_new = 0,
        dxf_tnst_correction = 1,
        dxf_tnst_cancel = 2


    ctypedef struct dxf_time_and_sale_t:
        dxf_event_flags_t event_flags,
        dxf_long_t index,
        dxf_long_t time,
        dxf_char_t exchange_code,
        dxf_double_t price,
        dxf_int_t size,
        dxf_double_t bid_price,
        dxf_double_t ask_price,
        dxf_const_string_t exchange_sale_conditions,
        dxf_int_t raw_flags,
        dxf_const_string_t buyer,
        dxf_const_string_t seller,
        dxf_order_side_t side,
        dxf_tns_type_t type,
        dxf_bool_t is_valid_tick,
        dxf_bool_t is_eth_trade,
        dxf_char_t trade_through_exempt,
        dxf_bool_t is_spread_leg


# /* Candle ------------------------------------------------------------------- */
    ctypedef struct dxf_candle_t:
        dxf_event_flags_t event_flags,
        dxf_long_t index,
        dxf_long_t time,
        dxf_int_t sequence,
        dxf_double_t count,
        dxf_double_t open,
        dxf_double_t high,
        dxf_double_t low,
        dxf_double_t close,
        dxf_double_t volume,
        dxf_double_t vwap,
        dxf_double_t bid_volume,
        dxf_double_t ask_volume,
        dxf_int_t open_interest,
        dxf_double_t imp_volatility


# /* Greeks ------------------------------------------------------------------- */
    ctypedef struct dxf_greeks_t:
        dxf_event_flags_t event_flags,
        dxf_long_t index,
        dxf_long_t time,
        dxf_double_t price,
        dxf_double_t volatility,
        dxf_double_t delta,
        dxf_double_t gamma,
        dxf_double_t theta,
        dxf_double_t rho,
        dxf_double_t vega

# /* TheoPrice ---------------------------------------------------------------- */
# /* Event and record are the same */
    ctypedef dx_theo_price_t dxf_theo_price_t

# /* Underlying --------------------------------------------------------------- */
# /* Event and record are the same */
    ctypedef dx_underlying_t dxf_underlying_t

# /* Series ------------------------------------------------------------------- */
    ctypedef struct dxf_series_t:
        dxf_event_flags_t event_flags,
        dxf_long_t index,
        dxf_long_t time,
        dxf_int_t sequence,
        dxf_dayid_t expiration,
        dxf_double_t volatility,
        dxf_double_t put_call_ratio,
        dxf_double_t forward_price,
        dxf_double_t dividend,
        dxf_double_t interest


    ctypedef struct dxf_configuration_t:
        dxf_int_t version,
        dxf_string_t object


# /* -------------------------------------------------------------------------- */
# /*
#  *	Event data constants
#  */
# /* -------------------------------------------------------------------------- */

    # ctypedef static dxf_const_string_t DXF_ORDER_COMPOSITE_BID_STR = L"COMPOSITE_BID"
    # ctypedef static dxf_const_string_t DXF_ORDER_COMPOSITE_ASK_STR = L"COMPOSITE_ASK"

    # cdef static dxf_const_string_t DXF_ORDER_COMPOSITE_BID_STR = "COMPOSITE_BID"
    # cdef static dxf_const_string_t DXF_ORDER_COMPOSITE_ASK_STR = "COMPOSITE_ASK"

    cdef dxf_const_string_t DXF_ORDER_COMPOSITE_BID_STR = "COMPOSITE_BID"
    cdef dxf_const_string_t DXF_ORDER_COMPOSITE_ASK_STR = "COMPOSITE_ASK"

# /* -------------------------------------------------------------------------- */
# /*
#  *	Event candle attributes
#  */
# /* -------------------------------------------------------------------------- */

#define DXF_CANDLE_EXCHANGE_CODE_COMPOSITE_ATTRIBUTE L'\0'
#define DXF_CANDLE_EXCHANGE_CODE_ATTRIBUTE_DEFAULT DXF_CANDLE_EXCHANGE_CODE_COMPOSITE_ATTRIBUTE
#define DXF_CANDLE_PERIOD_VALUE_ATTRIBUTE_DEFAULT 1.0

    ctypedef enum dxf_candle_price_attribute_t:
        dxf_cpa_last,
        dxf_cpa_bid,
        dxf_cpa_ask,
        dxf_cpa_mark,
        dxf_cpa_settlement,

        dxf_cpa_count,

        dxf_cpa_default = dxf_cpa_last


    ctypedef enum dxf_candle_session_attribute_t:
        dxf_csa_any,
        dxf_csa_regular,

        dxf_csa_count,

        dxf_csa_default = dxf_csa_any


    ctypedef enum dxf_candle_type_period_attribute_t:
        dxf_ctpa_tick,
        dxf_ctpa_second,
        dxf_ctpa_minute,
        dxf_ctpa_hour,
        dxf_ctpa_day,
        dxf_ctpa_week,
        dxf_ctpa_month,
        dxf_ctpa_optexp,
        dxf_ctpa_year,
        dxf_ctpa_volume,
        dxf_ctpa_price,
        dxf_ctpa_price_momentum,
        dxf_ctpa_price_renko,

        dxf_ctpa_count,

        dxf_ctpa_default = dxf_ctpa_tick


    ctypedef enum dxf_candle_alignment_attribute_t:
        dxf_caa_midnight,
        dxf_caa_session,

        dxf_caa_count,

        dxf_caa_default = dxf_caa_midnight


# /* -------------------------------------------------------------------------- */
# /*
#  *	Events flag constants
#  */
# /* -------------------------------------------------------------------------- */

    ctypedef enum dxf_event_flag:
        dxf_ef_tx_pending = 0x01,
        dxf_ef_remove_event = 0x02,
        dxf_ef_snapshot_begin = 0x04,
        dxf_ef_snapshot_end = 0x08,
        dxf_ef_snapshot_snip = 0x10,
        dxf_ef_remove_symbol = 0x20


# /* -------------------------------------------------------------------------- */
# /*
# *   Additional event params struct
# */
# /* -------------------------------------------------------------------------- */

    ctypedef dxf_ulong_t dxf_time_int_field_t

    ctypedef struct dxf_event_params_t:
        dxf_event_flags_t flags,
        dxf_time_int_field_t time_int_field,
        dxf_ulong_t snapshot_key


# /* -------------------------------------------------------------------------- */
# /*
#  *	Event listener prototype
#
#  *  event type here is a one-bit mask, not an integer
#  *  from dx_eid_begin to dx_eid_count
#  */
# /* -------------------------------------------------------------------------- */


    ctypedef void (*dxf_event_listener_t) (int event_type, dxf_const_string_t symbol_name,
                                          const dxf_event_data_t* data, int data_count,
                                          void* user_data)

    ctypedef void (*dxf_event_listener_v2_t) (int event_type, dxf_const_string_t symbol_name,
                                          const dxf_event_data_t* data, int data_count,
                                          const dxf_event_params_t* event_params, void* user_data)

# /* -------------------------------------------------------------------------- */
# /*
#  *	Various event functions
#  */
# /* -------------------------------------------------------------------------- */
cdef extern from "EventData.h":
    cdef dxf_const_string_t dx_event_type_to_string (int event_type)
    cdef int dx_get_event_data_struct_size (int event_id)
    cdef dx_event_id_t dx_get_event_id_by_bitmask (int event_bitmask)

# /* -------------------------------------------------------------------------- */
# /*
#  *	Event subscription stuff
#  */
# /* -------------------------------------------------------------------------- */

    enum dx_subscription_type_t:
        dx_st_begin = 0,

        dx_st_ticker = dx_st_begin,
        dx_st_stream,
        dx_st_history,

        # /* add new subscription types above this line */

        dx_st_count


    struct dx_event_subscription_param_t:
        dx_record_id_t record_id,
        dx_subscription_type_t subscription_type


    struct dx_event_subscription_param_list_t:
        dx_event_subscription_param_t* elements,
        size_t size,
        size_t capacity


# /*
#  * Returns the list of subscription params. Fills records list according to event_id.
#  *
#  * You need to call dx_free(params.elements) to free resources.
#  */
    cdef size_t dx_get_event_subscription_params(dxf_connection_t connection, dx_order_source_array_ptr_t order_source, dx_event_id_t event_id,
                                        dxf_uint_t subscr_flags, dx_event_subscription_param_list_t* params)

# /* -------------------------------------------------------------------------- */
# /*
# *  Snapshot data structs
# */
# /* -------------------------------------------------------------------------- */
    #Not sure if it will go
    struct dxf_snapshot_data_t:
        int event_type,
        dxf_string_t symbol,

        size_t records_count,
        const dxf_event_data_t* records

    ctypedef dxf_snapshot_data_t* dxf_snapshot_data_ptr_t

#
# /* -------------------------------------------------------------------------- */
# /*
# *  Snapshot listener prototype
#
# *  snapshot_data - pointer to the received snapshot data
# *  user_data     - pointer to user struct, use NULL by default
# */
# /* -------------------------------------------------------------------------- */
#
    ctypedef void(*dxf_snapshot_listener_t) (const dxf_snapshot_data_ptr_t snapshot_data, void* user_data)
#
# /* -------------------------------------------------------------------------- */
# /*
# *  Incremental Snapshot listener prototype
#
# *  snapshot_data - pointer to the received snapshot data
# *  new_snapshot  - flag, is this call with new snapshot or incremental update.
# *  user_data     - pointer to user struct, use NULL by default
# */
# /* -------------------------------------------------------------------------- */
# #define DXF_IS_CANDLE_REMOVAL(c) (((c)->event_flags & dxf_ef_remove_event) != 0)
# #define DXF_IS_ORDER_REMOVAL(o) ((((o)->event_flags & dxf_ef_remove_event) != 0) || ((o)->size == 0))
# #define DXF_IS_SPREAD_ORDER_REMOVAL(o) ((((o)->event_flags & dxf_ef_remove_event) != 0) || ((o)->size == 0))
# #define DXF_IS_TIME_AND_SALE_REMOVAL(t) (((t)->event_flags & dxf_ef_remove_event) != 0)
# #define DXF_IS_GREEKS_REMOVAL(g) (((g)->event_flags & dxf_ef_remove_event) != 0)
# #define DXF_IS_SERIES_REMOVAL(s) (((s)->event_flags & dxf_ef_remove_event) != 0)
    ctypedef void(*dxf_snapshot_inc_listener_t) (const dxf_snapshot_data_ptr_t snapshot_data, int new_snapshot, void* user_data)
#
# /* -------------------------------------------------------------------------- */
# /*
# *  Price Level data structs
# */
# /* -------------------------------------------------------------------------- */
    struct dxf_price_level_element_t:
        dxf_double_t price
        dxf_long_t size
        dxf_long_t time



    struct dxf_price_level_book_data_t:
        dxf_const_string_t symbol

        size_t bids_count
        const dxf_price_level_element_t *bids

        size_t asks_count
        const dxf_price_level_element_t *asks

    ctypedef dxf_price_level_book_data_t* dxf_price_level_book_data_ptr_t

# /* -------------------------------------------------------------------------- */
# /*
# *  Price Level listener prototype
#
# *  book          - pointer to the received price book data.
# *                  bids and asks are sorted by price,
# *                  best bid (with largest price) and best ask
# *                  (with smallest price) are first elements
# *                  of corresponding arrays.
# *  user_data     - pointer to user struct, use NULL by default
# */
# /* -------------------------------------------------------------------------- */
#
    ctypedef void(*dxf_price_level_book_listener_t) (const dxf_price_level_book_data_ptr_t book, void* user_data)
#
# /* -------------------------------------------------------------------------- */
# /*
# *  Price Level listener prototype
#
# *  quote         - pointer to the received regional quote
# *  user_data     - pointer to user struct, use NULL by default
# */
# /* -------------------------------------------------------------------------- */
#
    ctypedef void(*dxf_regional_quote_listener_t) (dxf_const_string_t symbol, const dxf_quote_t* quotes, int count, void* user_data)
#
# /* -------------------------------------------------------------------------- */
# /*
#  *	Event data navigation functions
#  */
# /* -------------------------------------------------------------------------- */
#
# #ifdef __cplusplus
#     extern "C"
# #endif
# const dxf_event_data_t dx_get_event_data_item (int event_mask, const dxf_event_data_t data, size_t index)
#
# #endif /* EVENT_DATA_H_INCLUDED */


from dxfeed.core.pxd_include.DXTypes cimport *

cdef extern from "DXErrorCodes.h":
    ctypedef enum dx_error_code_t:
    # # /* common error codes */
    
        dx_ec_success = 0
        
        dx_ec_error_subsystem_failure  # # /* this code may be set if passed from another thread */
        
        dx_ec_invalid_func_param  # # /* the invalid param is submitted by client */
        dx_ec_invalid_func_param_internal  # # /* the invalid param is a result of internal error */
        
        dx_ec_internal_assert_violation
        
        # # /* memory error codes */
        
        dx_mec_insufficient_memory
        
        # # /* socket error codes */
        
        dx_sec_socket_subsystem_init_failed,  # # /* Win32-specific */
        dx_sec_socket_subsystem_init_required,  # # /* Win32-specific */
        dx_sec_socket_subsystem_incompatible_version,  # # /* Win32-specific */
        dx_sec_connection_gracefully_closed
        dx_sec_network_is_down
        dx_sec_blocking_call_in_progress
        dx_sec_addr_family_not_supported
        dx_sec_no_sockets_available
        dx_sec_no_buffer_space_available
        dx_sec_proto_not_supported
        dx_sec_socket_type_proto_incompat
        dx_sec_socket_type_addrfam_incompat
        dx_sec_addr_already_in_use
        dx_sec_blocking_call_interrupted
        dx_sec_nonblocking_oper_pending
        dx_sec_addr_not_valid
        dx_sec_connection_refused
        dx_sec_invalid_ptr_arg
        dx_sec_invalid_arg
        dx_sec_sock_already_connected
        dx_sec_network_is_unreachable
        dx_sec_sock_oper_on_nonsocket
        dx_sec_connection_timed_out
        dx_sec_res_temporarily_unavail
        dx_sec_permission_denied
        dx_sec_network_dropped_connection
        dx_sec_socket_not_connected
        dx_sec_operation_not_supported
        dx_sec_socket_shutdown
        dx_sec_message_too_long
        dx_sec_no_route_to_host
        dx_sec_connection_aborted
        dx_sec_connection_reset
        dx_sec_persistent_temp_error
        dx_sec_unrecoverable_error
        dx_sec_not_enough_memory
        dx_sec_no_data_on_host
        dx_sec_host_not_found
    
        dx_sec_generic_error
        
        # /* thread error codes */
        
        dx_tec_not_enough_sys_resources
        dx_tec_permission_denied
        dx_tec_invalid_res_operation
        dx_tec_invalid_resource_id
        dx_tec_deadlock_detected
        dx_tec_not_enough_memory
        dx_tec_resource_busy
    
        dx_tec_generic_error
        
        # /* network error codes */
        
        dx_nec_invalid_port_value
        dx_nec_invalid_function_arg
        dx_nec_connection_closed
        dx_nec_open_connection_error
        dx_nec_unknown_codec
        
        # /* buffered I/O error codes */
        
        dx_bioec_buffer_overflow
        dx_bioec_buffer_not_initialized
        dx_bioec_index_out_of_bounds
        dx_bioec_buffer_underflow
        
        # /* UTF error codes */
        
        dx_utfec_bad_utf_data_format
        dx_utfec_bad_utf_data_format_server
        
        # /* penta codec error codes */
        
        dx_pcec_reserved_bit_sequence
        dx_pcec_invalid_symbol_length
        dx_pcec_invalid_event_flag
        
        # /* event subscription error codes */
        
        dx_esec_invalid_event_type
        dx_esec_invalid_subscr_id
        dx_esec_invalid_symbol_name
        dx_esec_invalid_listener
        
        # /* logger error codes */
        
        dx_lec_failed_to_open_file
        
        # /* protocol message error codes */
        
        dx_pmec_invalid_message_type
        
        # /* protocol error codes */
        
        dx_pec_unexpected_message_type
        dx_pec_unexpected_message_type_internal
        dx_pec_descr_record_field_info_corrupted
        dx_pec_message_incomplete
        dx_pec_invalid_message_length
        dx_pec_server_message_not_supported
        dx_pec_invalid_symbol
        dx_pec_record_description_not_received
        dx_pec_record_field_type_not_supported
        dx_pec_record_info_corrupted
        dx_pec_unknown_record_name
        dx_pec_record_not_supported
        dx_pec_describe_protocol_message_corrupted
        dx_pec_unexpected_message_sequence_internal
        dx_pec_local_message_not_supported_by_server
        dx_pec_inconsistent_message_support
        dx_pec_authentication_error
        dx_pec_credentials_required
        
        # /* connection error codes */
        
        dx_cec_invalid_connection_handle
        dx_cec_invalid_connection_handle_internal
        dx_cec_connection_context_not_initialized
        dx_cec_invalid_connection_context_subsystem_id
    
        # /* candle event error codes*/
    
        dx_ceec_invalid_candle_period_value
    
    
        # /* snapshot error codes */
    
        dx_ssec_invalid_snapshot_id
        dx_ssec_invalid_event_id
        dx_ssec_invalid_symbol
        dx_ssec_snapshot_exist
        dx_ssec_invalid_listener
        dx_ssec_unknown_state
        dx_ssec_duplicate_record
    
        # /* configuration record serialization deserialization error codes */
    
        dx_csdec_protocol_error
        dx_csdec_unsupported_version
        
        # /* miscellaneous error codes */
        
        # /* error code count */
        # /* this MUST be the last element in the enumeration */
        
        dx_ec_count


# # /* -------------------------------------------------------------------------- */
# # /*
#  *	Message description functions
#  */
# # /* -------------------------------------------------------------------------- */

cdef dxf_const_string_t dx_get_error_description (dx_error_code_t code)
    


# include "DXTypes.pxd"
from dxfeed.core.pxd_include.DXTypes cimport *
from dxfeed.core.pxd_include.EventData cimport *
# /* -------------------------------------------------------------------------- */
# /*
#  *	DXFeed API generic types
#  */
# /* -------------------------------------------------------------------------- */


cdef extern from "DXFeed.h":
    cdef int DXF_SUCCESS = 1
    cdef int DXF_FAILURE = 0
 
# /* -------------------------------------------------------------------------- */
# /*
#  *	DXFeed API function return value codes
#  */
# /* -------------------------------------------------------------------------- */
# 
# #define DXF_SUCCESS 1
# #define DXF_FAILURE 0

# /* -------------------------------------------------------------------------- */
# /*
#  *	DXFeed API generic types
#  */
# /* -------------------------------------------------------------------------- */

    ctypedef void (*dxf_conn_termination_notifier_t) (dxf_connection_t connection, void* user_data)
    ctypedef void (*dxf_conn_status_notifier_t) (dxf_connection_t connection,
                                            dxf_connection_status_t old_status,
                                            dxf_connection_status_t new_status,
                                            void* user_data)

# /* the low level callback types, required in case some thread-specific initialization must be performed
#    on the client side on the thread creation/destruction */

    ctypedef int (*dxf_socket_thread_creation_notifier_t) (dxf_connection_t connection, void* user_data)
    ctypedef void (*dxf_socket_thread_destruction_notifier_t) (dxf_connection_t connection, void* user_data)

# /* -------------------------------------------------------------------------- */
# /*
#  *	DXFeed C API functions
#  
#  *  All functions return DXF_SUCCESS on success and DXF_FAILURE if some error
#  *  has occurred. Use 'dxf_get_last_error' to retrieve the error code
#  *  and description.
#  */
# /* -------------------------------------------------------------------------- */

# /*
#  *	Creates connection with the specified parameters.
#  
#  *  address - the single address: "host:port" or just "host"
#  *            address with credentials: "host:port[username=xxx,password=yyy]"
#  *            multiple addresses: "(host1:port1)(host2)(host3:port3[username=xxx,password=yyy])"
#  *            the data from file: "/path/to/file" on *nix and "drive:\path\to\file" on Windows
#  *  notifier - the callback to inform the client side that the connection has stumbled upon and error and will go reconnecting
#  *  conn_status_notifier - the callback to inform the client side that the connection status has changed
#  *  stcn - the callback for informing the client side about the socket thread creation
#            may be set to NULL if no specific action is required to perform on the client side on a new thread creation
#  *  shdn - the callback for informing the client side about the socket thread destruction
#            may be set to NULL if no specific action is required to perform on the client side on a thread destruction
#  *  user_data - the user defined value passed to the termination notifier callback along with the connection handle may be set
#                 to whatever value
#  *  OUT connection - the handle of the created connection
#  */
    ERRORCODE dxf_create_connection (const char* address,
                                            dxf_conn_termination_notifier_t notifier,
                                            dxf_conn_status_notifier_t conn_status_notifier,
                                            dxf_socket_thread_creation_notifier_t stcn,
                                            dxf_socket_thread_destruction_notifier_t stdn,
                                            void* user_data,
                                            dxf_connection_t* connection)

# /*
#  *	Creates connection with the specified parameters and basic authorization.
#
#  *  address - the single address: "host:port" or just "host"
#  *            address with credentials: "host:port[username=xxx,password=yyy]"
#  *            multiple addresses: "(host1:port1)(host2)(host3:port3[username=xxx,password=yyy])"
#  *            the data from file: "/path/to/file" on *nix and "drive:\path\to\file" on Windows
#  *  user - the user name
#  *  password - the user password
#  *  notifier - the callback to inform the client side that the connection has stumbled upon and error and will go reconnecting
#  *  conn_status_notifier - the callback to inform the client side that the connection status has changed
#  *  stcn - the callback for informing the client side about the socket thread creation
#            may be set to NULL if no specific action is required to perform on the client side on a new thread creation
#  *  shdn - the callback for informing the client side about the socket thread destruction
#            may be set to NULL if no specific action is required to perform on the client side on a thread destruction
#  *  user_data - the user defined value passed to the termination notifier callback along with the connection handle may be set
#                 to whatever value
#  *  OUT connection - the handle of the created connection.
#  *
#  * Note: the user and password data from argument have a higher priority than address credentials.
#  */
    ERRORCODE dxf_create_connection_auth_basic(const char* address,
                                                      const char* user,
                                                      const char* password,
                                                      dxf_conn_termination_notifier_t notifier,
                                                      dxf_conn_status_notifier_t conn_status_notifier,
                                                      dxf_socket_thread_creation_notifier_t stcn,
                                                      dxf_socket_thread_destruction_notifier_t stdn,
                                                      void* user_data,
                                                      dxf_connection_t* connection)

# /*
#  *	Creates connection with the specified parameters and token authorization.
#
#  *  address - the single address: "host:port" or just "host"
#  *            address with credentials: "host:port[username=xxx,password=yyy]"
#  *            multiple addresses: "(host1:port1)(host2)(host3:port3[username=xxx,password=yyy])"
#  *            the data from file: "/path/to/file" on *nix and "drive:\path\to\file" on Windows
#  *  token - the authorization token
#  *  notifier - the callback to inform the client side that the connection has stumbled upon and error and will go reconnecting
#  *  conn_status_notifier - the callback to inform the client side that the connection status has changed
#  *  stcn - the callback for informing the client side about the socket thread creation
#            may be set to NULL if no specific action is required to perform on the client side on a new thread creation
#  *  shdn - the callback for informing the client side about the socket thread destruction
#            may be set to NULL if no specific action is required to perform on the client side on a thread destruction
#  *  user_data - the user defined value passed to the termination notifier callback along with the connection handle may be set
#                 to whatever value
#  *  OUT connection - the handle of the created connection.
#  *
#  * Note: the token data from argument have a higher priority than address credentials.
#  */
    ERRORCODE dxf_create_connection_auth_bearer(const char* address,
                                                const char* token,
                                                dxf_conn_termination_notifier_t notifier,
                                                dxf_conn_status_notifier_t conn_status_notifier,
                                                dxf_socket_thread_creation_notifier_t stcn,
                                                dxf_socket_thread_destruction_notifier_t stdn,
                                                void* user_data,
                                                dxf_connection_t* connection)

# /*
#  *	Creates connection with the specified parameters and custom described authorization.
#
#  *  address - the single address: "host:port" or just "host"
#  *            address with credentials: "host:port[username=xxx,password=yyy]"
#  *            multiple addresses: "(host1:port1)(host2)(host3:port3[username=xxx,password=yyy])"
#  *            the data from file: "/path/to/file" on *nix and "drive:\path\to\file" on Windows
#  *  authscheme - the authorization scheme
#  *  authdata - the authorization data
#  *  notifier - the callback to inform the client side that the connection has stumbled upon and error and will go reconnecting
#  *  conn_status_notifier - the callback to inform the client side that the connection status has changed
#  *  stcn - the callback for informing the client side about the socket thread creation
#            may be set to NULL if no specific action is required to perform on the client side on a new thread creation
#  *  shdn - the callback for informing the client side about the socket thread destruction
#            may be set to NULL if no specific action is required to perform on the client side on a thread destruction
#  *  user_data - the user defined value passed to the termination notifier callback along with the connection handle may be set
#                 to whatever value
#  *  OUT connection - the handle of the created connection.
#  *
#  * Note: the authscheme and authdata from argument have a higher priority than address credentials.
#  */
    ERRORCODE dxf_create_connection_auth_custom(const char* address,
                                                       const char* authscheme,
                                                       const char* authdata,
                                                       dxf_conn_termination_notifier_t notifier,
                                                       dxf_conn_status_notifier_t conn_status_notifier,
                                                       dxf_socket_thread_creation_notifier_t stcn,
                                                       dxf_socket_thread_destruction_notifier_t stdn,
                                                       void* user_data,
                                                       dxf_connection_t* connection)

# /*
#  *	Closes a connection.
#
#  *  connection - a handle of a previously created connection
#  */
    ERRORCODE dxf_close_connection (dxf_connection_t connection)

# /*
#  *	Creates a subscription with the specified parameters.
#
#  *  connection - a handle of a previously created connection which the subscription will be using
#  *  event_types - a bitmask of the subscription event types. See 'dx_event_id_t' and 'DX_EVENT_BIT_MASK'
#  *                for information on how to create an event type bitmask
#  *  OUT subscription - a handle of the created subscription
#  */
    ERRORCODE dxf_create_subscription (dxf_connection_t connection, int event_types,
                                              dxf_subscription_t* subscription)

# /*
#  *	Creates a subscription with the specified parameters.
#
#  *  connection - a handle of a previously created connection which the subscription will be using
#  *  event_types - a bitmask of the subscription event types. See 'dx_event_id_t' and 'DX_EVENT_BIT_MASK'
#  *                for information on how to create an event type bitmask
#  *  time - UTC time in the past (unix time in milliseconds)
#  *  OUT subscription - a handle of the created subscription
#  */
    ERRORCODE dxf_create_subscription_timed(dxf_connection_t connection, int event_types, 
                                                   dxf_long_t time,
                                                   dxf_subscription_t* subscription)

# /*
#  *	Closes a subscription.
#  *  All the data associated with it will be freed.
#
#  *  subscription - a handle of the subscription to close
#  */
    ERRORCODE dxf_close_subscription (dxf_subscription_t subscription)

# /*
#  *	Adds a single symbol to the subscription.
#
#  *  subscription - a handle of the subscription to which a symbol is added
#  *  symbol - the symbol to add
#  */
    ERRORCODE dxf_add_symbol (dxf_subscription_t subscription, dxf_const_string_t symbol)

# /*
#  *	Adds several symbols to the subscription.
#  *  No error occurs if the symbol is attempted to add for the second time.
#
#  *  subscription - a handle of the subscription to which the symbols are added
#  *  symbols - the symbols to add
#  *  symbol_count - a number of symbols
#  */
    ERRORCODE dxf_add_symbols (dxf_subscription_t subscription, dxf_const_string_t* symbols, int symbol_count)

# /*
#  *	Adds a candle symbol to the subscription.
#
#  *  subscription - a handle of the subscription to which a symbol is added
#  *  candle_attributes - pointer to the candle struct
#  */
    ERRORCODE dxf_add_candle_symbol(dxf_subscription_t subscription, dxf_candle_attributes_t candle_attributes)

# /*
# *	Remove a candle symbol from the subscription.
#
# *  subscription - a handle of the subscription from symbol will be removed
# *  candle_attributes - pointer to the candle struct
# */
    ERRORCODE dxf_remove_candle_symbol(dxf_subscription_t subscription, dxf_candle_attributes_t candle_attributes)

# /*
#  *	Removes a single symbol from the subscription.
#
#  *  subscription - a handle of the subscription from which a symbol is removed
#  *  symbol - the symbol to remove
#  */
    ERRORCODE dxf_remove_symbol (dxf_subscription_t subscription, dxf_const_string_t symbol)

# /*
#  *	Removes several symbols from the subscription.
#  *  No error occurs if it's attempted to remove symbols that weren't added beforehand.
#
#  *  subscription - a handle of the subscription to which the symbols are added
#  *  symbols - the symbols to remove
#  *  symbol_count - a number of symbols
#  */
    ERRORCODE dxf_remove_symbols (dxf_subscription_t subscription, dxf_const_string_t* symbols, int symbol_count)

# /*
#  *	Retrieves the list of symbols currently added to the subscription.
#  *  The memory for the resulting list is allocated internally, so no actions to free it are required.
#  *  The symbol name buffer is guaranteed to be valid until either the subscription symbol list is changed or a new call
#  *  of this function is performed.
#
#  *  subscription - a handle of the subscriptions whose symbols are to retrieve
#  *  OUT symbols - a pointer to the string array object to which the symbol list is to be stored
#  *  OUT symbol_count - a pointer to the variable to which the symbol count is to be stored
#  */
    ERRORCODE dxf_get_symbols (dxf_subscription_t subscription, dxf_const_string_t** symbols, int* symbol_count)

# /*
#  *	Sets the symbols for the subscription.
#  *  The difference between this function and 'dxf_add_symbols' is that all the previously added symbols
#  *  that do not belong to the symbol list passed to this function will be removed.
#
#  *  subscription - a handle of the subscription whose symbols are to be set
#  *  symbols - the symbol list to set
#  *  symbol_count - the symbol count
#  */
    ERRORCODE dxf_set_symbols (dxf_subscription_t subscription, dxf_const_string_t* symbols, int symbol_count)

# /*
#  *	Removes all the symbols from the subscription.
#
#  *  subscription - a handle of the subscription whose symbols are to be cleared
#  */
    ERRORCODE dxf_clear_symbols (dxf_subscription_t subscription)

# /*
#  *	Attaches a listener callback to the subscription.
#  *  This callback will be invoked when the new event data for the subscription symbols arrives.
#  *  No error occurs if it's attempted to attach the same listener twice or more.
#
#  *  subscription - a handle of the subscription to which a listener is to be attached
#  *  event_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_attach_event_listener (dxf_subscription_t subscription, dxf_event_listener_t event_listener,
                                                void* user_data)

# /*
#  *	Detaches a listener from the subscription.
#  *  No error occurs if it's attempted to detach a listener which wasn't previously attached.
#
#  *  subscription - a handle of the subscription from which a listener is to be detached
#  *  event_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_detach_event_listener (dxf_subscription_t subscription, dxf_event_listener_t event_listener)

# /*
# *  Attaches a extended listener callback to the subscription.
# *  This callback will be invoked when the new event data for the subscription symbols arrives.
# *  No error occurs if it's attempted to attach the same listener twice or more.
# *
# *  This listener differs with extend number of callback parameters.
# *
# *  subscription - a handle of the subscription to which a listener is to be attached
# *  event_listener - a listener callback function pointer
# *  user_data - if there isn't user data pass NULL
# */
    ERRORCODE dxf_attach_event_listener_v2(dxf_subscription_t subscription, 
                                                  dxf_event_listener_v2_t event_listener,
                                                  void* user_data)

# /*
# *  Detaches a extended listener from the subscription.
# *  No error occurs if it's attempted to detach a listener which wasn't previously attached.
#
# *  subscription - a handle of the subscription from which a listener is to be detached
# *  event_listener - a listener callback function pointer
# */
    ERRORCODE dxf_detach_event_listener_v2(dxf_subscription_t subscription, 
                                                  dxf_event_listener_v2_t event_listener)

# /*
#  *	Retrieves the subscription event types.
#
#  *  subscription - a handle of the subscription whose event types are to be retrieved
#  *  OUT event_types - a pointer to the variable to which the subscription event types bitmask is to be stored
#  */
    ERRORCODE dxf_get_subscription_event_types (dxf_subscription_t subscription, int* event_types)

# /*
#  *	Retrieves the last event data of the specified symbol and type for the connection.
#
#  *  connection - a handle of the connection whose data is to be retrieved
#  *  event_type - an event type bitmask defining a single event type
#  *  symbol - a symbol name
#  *  OUT event_data - a pointer to the variable to which the last data buffer pointer is stored if there was no
#                      data for this connection/event type/symbol, NULL will be stored
#  */
    ERRORCODE dxf_get_last_event (dxf_connection_t connection, int event_type, dxf_const_string_t symbol,
                                         dxf_event_data_t* event_data)
# /*
#  *	Retrieves the last error info.
#  *  The error is stored on per-thread basis. If the connection termination notifier callback was invoked, then
#  *  to retrieve the connection's error code call this function right from the callback function.
#  *  If an error occurred within the error storage subsystem itself, the function will return DXF_FAILURE.
#
#  *  OUT error_code - a pointer to the variable where the error code is to be stored
#  *  OUT error_descr - a pointer to the variable where the human-friendly error description is to be stored
#                       may be NULL if the text representation of error is not required
#  */
    ERRORCODE dxf_get_last_error (int* error_code, dxf_const_string_t* error_descr)

# /*
#  *	Initializes the internal logger.
#  *  Various actions and events, including the errors, are being logged throughout the library. They may be stored
#  *  into the file.
#  *
#  *  file_name - a full path to the file where the log is to be stored
#  *  rewrite_file - a flag defining the file open mode if it's nonzero then the log file will be rewritten
#  *  show_timezone_info - a flag defining the time display option in the log file if it's nonzero then
#                          the time will be displayed with the timezone suffix
#  *  verbose - a flag defining the logging mode if it's nonzero then the verbose logging will be enabled
#  */
    ERRORCODE dxf_initialize_logger (const char* file_name, int rewrite_file, int show_timezone_info, int verbose)

# /*
#  *  Clear current sources and add new one to subscription
#  *  Warning: you must configure order source before dxf_add_symbols/dxf_add_symbol call
#  *
#  *  subscription - a handle of the subscription where source will be changed
#  *  source - source of order to set, 4 symbols maximum length
#  */
    ERRORCODE dxf_set_order_source(dxf_subscription_t subscription, const char* source)

# /*
#  *  Add a new source to subscription
#  *  Warning: you must configure order source before dxf_add_symbols/dxf_add_symbol call
#  *
#  *  subscription - a handle of the subscription where source will be changed
#  *  source - source of order event to add, 4 symbols maximum length
#  */
    ERRORCODE dxf_add_order_source(dxf_subscription_t subscription, const char* source)

# /*
#  *	API that allows user to create candle symbol attributes
#  *
#  *  base_symbol - the symbols to add
#  *  exchange_code - exchange attribute of this symbol
#  *  period_value -  aggregation period value of this symbol
#  *  period_type - aggregation period type of this symbol
#  *  price - price type attribute of this symbol
#  *  session - session attribute of this symbol
#  *  alignment - alignment attribute of this symbol
#  *  candle_attributes - pointer to the configured candle attributes struct
#  */
    ERRORCODE dxf_create_candle_symbol_attributes(dxf_const_string_t base_symbol,
                                                         dxf_char_t exchange_code,
                                                         dxf_double_t period_value,
                                                         dxf_candle_type_period_attribute_t period_type,
                                                         dxf_candle_price_attribute_t price,
                                                         dxf_candle_session_attribute_t session,
                                                         dxf_candle_alignment_attribute_t alignment,
                                                         dxf_candle_attributes_t* candle_attributes)

# /*
#  *	Free memory allocated by dxf_initialize_candle_symbol_attributes(...) function
#  *
#  *  candle_attributes - pointer to the candle attributes struct
#  */
    ERRORCODE dxf_delete_candle_symbol_attributes(dxf_candle_attributes_t candle_attributes)

# /*
#  *  Creates snapshot with the specified parameters.
#  *
#  *  For Order or Candle events (dx_eid_order or dx_eid_candle) please use
#  *  short form of this function: dxf_create_order_snapshot or dxf_create_candle_snapshot
#  *  respectively.
#  *
#  *  For order events (event_id is 'dx_eid_order')
#  *  If source is NULL string subscription on Order event will be performed. You can specify order
#  *  source for Order event by passing suffix: "BYX", "BZX", "DEA", "DEX", "ISE", "IST", "NTV".
#  *  If source is equal to "COMPOSITE_BID" or "COMPOSITE_ASK" subscription on MarketMaker event will
#  *  be performed. For other events source parameter does not matter.
#  *
#  *  connection - a handle of a previously created connection which the subscription will be using
#  *  event_id - single event id. Next events is supported: dxf_eid_order, dxf_eid_candle,
#                dx_eid_spread_order, dx_eid_time_and_sale, dx_eid_greeks, dx_eid_series.
#  *  symbol - the symbol to add.
#  *  source - order source for Order, which can be one of following: "BYX", "BZX", "DEA", "DEX",
#  *           "ISE", "IST", "NTV". For MarketMaker subscription use "COMPOSITE_BID" or
#  *           "COMPOSITE_ASK" keyword.
#  *  time - time in the past (unix time in milliseconds).
#  *  OUT snapshot - a handle of the created snapshot
#  */
    ERRORCODE dxf_create_snapshot(dxf_connection_t connection, dx_event_id_t event_id,
                                         dxf_const_string_t symbol, const char* source, 
                                         dxf_long_t time, dxf_snapshot_t* snapshot)

# /*
#  *  Creates Order snapshot with the specified parameters.
#  *
#  *  If source is NULL string subscription on Order event will be performed. You can specify order
#  *  source for Order event by passing suffix: "BYX", "BZX", "DEA", "DEX", "ISE", "IST", "NTV".
#  *  If source is equal to "COMPOSITE_BID" or "COMPOSITE_ASK" subscription on MarketMaker event will
#  *  be performed.
#  *
#  *  connection - a handle of a previously created connection which the subscription will be using
#  *  symbol - the symbol to add
#  *  source - order source for Order, which can be one of following: "BYX", "BZX", "DEA", "DEX",
#  *           "ISE", "IST", "NTV". For MarketMaker subscription use "COMPOSITE_BID" or
#  *           "COMPOSITE_ASK" keyword.
#  *  time - time in the past (unix time in milliseconds)
#  *  OUT snapshot - a handle of the created snapshot
#  */
    ERRORCODE dxf_create_order_snapshot(dxf_connection_t connection, 
                                               dxf_const_string_t symbol, const char* source,
                                               dxf_long_t time, dxf_snapshot_t* snapshot)

# /*
#  *  Creates Candle snapshot with the specified parameters.
#  *
#  *  connection - a handle of a previously created connection which the subscription will be using
#  *  candle_attributes - object specified symbol attributes of candle
#  *  time - time in the past (unix time in milliseconds)
#  *  OUT snapshot - a handle of the created snapshot
#  */
    ERRORCODE dxf_create_candle_snapshot(dxf_connection_t connection, 
                                                dxf_candle_attributes_t candle_attributes, 
                                                dxf_long_t time, dxf_snapshot_t* snapshot)

# /*
#  *  Closes a snapshot.
#  *  All the data associated with it will be freed.
#  *
#  *  snapshot - a handle of the snapshot to close
#  */
    ERRORCODE dxf_close_snapshot(dxf_snapshot_t snapshot)

# /*
#  *  Attaches a listener callback to the snapshot.
#  *  This callback will be invoked when the new snapshot arrives or existing updates.
#  *  No error occurs if it's attempted to attach the same listener twice or more.
#  *
#  *  snapshot - a handle of the snapshot to which a listener is to be attached
#  *  snapshot_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_attach_snapshot_listener(dxf_snapshot_t snapshot, 
                                                  dxf_snapshot_listener_t snapshot_listener,
                                                  void* user_data)

# /*
#  *  Detaches a listener from the snapshot.
#  *  No error occurs if it's attempted to detach a listener which wasn't previously attached.
#  *
#  *  snapshot - a handle of the snapshot to which a listener is to be detached
#  *  snapshot_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_detach_snapshot_listener(dxf_snapshot_t snapshot, 
                                                  dxf_snapshot_listener_t snapshot_listener)

# /*
#  *  Attaches a incremental listener callback to the snapshot.
#  *  This callback will be invoked when the new snapshot arrives or existing updates.
#  *  No error occurs if it's attempted to attach the same listener twice or more.
#  *
#  *  snapshot - a handle of the snapshot to which a listener is to be attached
#  *  snapshot_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_attach_snapshot_inc_listener(dxf_snapshot_t snapshot, 
                                                  dxf_snapshot_inc_listener_t snapshot_listener,
                                                  void* user_data)

# /*
#  *  Detaches a listener from the snapshot.
#  *  No error occurs if it's attempted to detach a listener which wasn't previously attached.
#  *
#  *  snapshot - a handle of the snapshot to which a listener is to be detached
#  *  snapshot_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_detach_snapshot_inc_listener(dxf_snapshot_t snapshot, 
                                                  dxf_snapshot_inc_listener_t snapshot_listener)

# /*
#  *  Retrieves the symbol currently added to the snapshot subscription.
#  *  The memory for the resulting symbol is allocated internally, so no actions to free it are required.
#  *
#  *  snapshot - a handle of the snapshot to which a listener is to be detached
#  *  OUT symbol - a pointer to the string object to which the symbol is to be stored
# */
    ERRORCODE dxf_get_snapshot_symbol(dxf_snapshot_t snapshot, dxf_string_t* symbol)

# /*
#  *  Creates Price Level book with the specified parameters.
#  *
#  *  connection - a handle of a previously created connection which the subscription will be using
#  *  symbol - the symbol to use
#  *  sources - order sources for Order, NULL-terminated list. Each element can be one of following:
#  *            "BYX", "BZX", "DEA", "DEX", "ISE", "IST", "NTV".
#  *  OUT book - a handle of the created price level book
#  */
    ERRORCODE dxf_create_price_level_book(dxf_connection_t connection, 
                                                 dxf_const_string_t symbol, const char** sources,
                                                 dxf_price_level_book_t* book)

# /*
#  *  Closes a price level book.
#  *  All the data associated with it will be freed.
#  *
#  *  book - a handle of the price level book to close
#  */
    ERRORCODE dxf_close_price_level_book(dxf_price_level_book_t book)

# /*
#  *  Attaches a listener callback to the price level book.
#  *  This callback will be invoked when price levels change.
#  *  No error occurs if it's attempted to attach the same listener twice or more.
#  *
#  *  book - a handle of the book to which a listener is to be detached
#  *  book_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_attach_price_level_book_listener(dxf_price_level_book_t book, 
                                                          dxf_price_level_book_listener_t book_listener,
                                                          void* user_data)

# /*
#  *  Detaches a listener from the snapshot.
#  *  No error occurs if it's attempted to detach a listener which wasn't previously attached.
#  *
#  *  book - a handle of the book to which a listener is to be detached
#  *  book_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_detach_price_level_book_listener(dxf_price_level_book_t book, 
                                                          dxf_price_level_book_listener_t book_listener)

# /*
#  *  Creates Regional book with the specified parameters.
#  *  Regional book is like Price Level Book but uses regional data instead of full depth order book.
#  *
#  *  connection - a handle of a previously created connection which the subscription will be using
#  *  symbol - the symbol to use
#  *  OUT book - a handle of the created regional book
#  */
    ERRORCODE dxf_create_regional_book(dxf_connection_t connection, 
                                              dxf_const_string_t symbol,
                                              dxf_regional_book_t* book)

# /*
#  *  Closes a regional book.
#  *  All the data associated with it will be freed.
#  *
#  *  book - a handle of the price level book to close
#  */
    ERRORCODE dxf_close_regional_book(dxf_regional_book_t book)

# /*
#  *  Attaches a listener callback to regional book.
#  *  This callback will be invoked when price levels created from regional data change.
#  *
#  *  book - a handle of the book to which a listener is to be detached
#  *  book_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_attach_regional_book_listener(dxf_regional_book_t book, 
                                                       dxf_price_level_book_listener_t book_listener,
                                                       void* user_data)

# /*
#  *  Detaches a listener from the regional book.
#  *  No error occurs if it's attempted to detach a listener which wasn't previously attached.
#  *
#  *  book - a handle of the book to which a listener is to be detached
#  *  book_listener - a listener callback function pointer
#  */
    ERRORCODE dxf_detach_regional_book_listener(dxf_regional_book_t book, 
                                                       dxf_price_level_book_listener_t book_listener)

# /*
# *  Attaches a listener callback to regional book.
# *  This callback will be invoked when new regional quotes are received.
# *
# *  book - a handle of the book to which a listener is to be detached
# *  listener - a listener callback function pointer
# */
    ERRORCODE dxf_attach_regional_book_listener_v2(dxf_regional_book_t book,
    dxf_regional_quote_listener_t listener, void* user_data)

# /*
# *  Detaches a listener from the regional book.
# *  No error occurs if it's attempted to detach a listener which wasn't previously attached.
# *
# *  book - a handle of the book to which a listener is to be detached
# *  book_listener - a listener callback function pointer
# */
    ERRORCODE dxf_detach_regional_book_listener_v2(dxf_regional_book_t book,
    dxf_regional_quote_listener_t listener)

# /*
#  *  Add dumping of incoming traffic into specific file
#  *
#  *  connection - a handle of a previously created connection which the subscription will be using
#  *  raw_file_name - raw data file name
#  */
    ERRORCODE dxf_write_raw_data(dxf_connection_t connection, const char* raw_file_name)

# /*
# *  Retrieves the array of key-value pairs (properties) for specified connection. The memory for the resulting array
# *  is allocated during execution of the function and SHOULD be free by caller with dxf_free_connection_properties_snapshot
# *  function. So done because connection properties can be changed during reconnection. Returned array is a snapshot
# *  of properties at the moment of the call.
# *
# *  connection - a handle of a previously created connection
# *  OUT properties - address of pointer to store address of key-value pairs array
# *  OUT count - address of variable to store length of key-value pairs array
# */
    ERRORCODE dxf_get_connection_properties_snapshot(dxf_connection_t connection,
                                                            dxf_property_item_t** properties,
                                                            int* count)

# /*
# * Frees memory allocated during dxf_get_connection_properties_snapshot function execution
# *
# *  properties - pointer to the key-value pairs array
# *  count - length of key-value pairs array
# */
    ERRORCODE dxf_free_connection_properties_snapshot(dxf_property_item_t* properties, int count)

# /*
# *  Retrieves the null-terminated string with current connected address in format <host>:<port>. If (*address)
# *  points to NULL then connection is not connected (reconnection, no valid addresses, closed connection and others).
# *  The memory for the resulting string is allocated during execution of the function and SHOULD be free by caller
# *  with call of dxf_free function. So done because inner string with connected address can be free during reconnection.
# *
# *  connection - a handle of a previously created connection
# *  OUT address - address of pointer to store address of the null-terminated string with current connected address
# */
    ERRORCODE dxf_get_current_connected_address(dxf_connection_t connection, char** address)

# /*
# *  Retrieves the current connection status
# *
# *  connection - a handle of a previously created connection
# *  OUT status - connection status
# */
    ERRORCODE dxf_get_current_connection_status(dxf_connection_t connection, dxf_connection_status_t* status)

# /*
# * Frees memory allocated in API functions from this module
# *
# *  pointer - pointer to memory allocated earlier in some API function from this module
# */
    ERRORCODE dxf_free(void* pointer)

#endif /*     _H_INCLUDED */



# distutils: language = c++
# cython: always_allow_keywords=True
from libcpp.deque cimport deque as cppdq

from dxfeed.core.utils.helpers cimport *
from dxfeed.core.utils.helpers import *
cimport dxfeed.core.pxd_include.DXFeed as clib
cimport dxfeed.core.pxd_include.DXErrorCodes as dxec
cimport dxfeed.core.listeners.listener as lis
from dxfeed.core.utils.data_class import DequeWithLock as deque_wl
from datetime import datetime
import pandas as pd
from typing import Optional, Union, Iterable
from warnings import warn

# for importing variables
import dxfeed.core.listeners.listener as lis
from dxfeed.core.pxd_include.EventData cimport *

cpdef int process_last_error(verbose: bool=True):
    """
    Function retrieves last error

    Parameters
    ----------
    verbose: bool
        If True error description is printed
        
    Returns
    -------
    error_code: int
        Error code is returned
    """
    cdef int error_code = dxec.dx_ec_success
    cdef dxf_const_string_t error_descr = NULL
    cdef int res

    res = clib.dxf_get_last_error(&error_code, &error_descr)

    if res == clib.DXF_SUCCESS:
        if error_code == dxec.dx_ec_success and verbose:
            print("no error information is stored")

        if verbose:
            print("Error occurred and successfully retrieved:\n",
                  f"error code = {error_code}, description = {unicode_from_dxf_const_string_t(error_descr)}")

    return error_code


cdef class ConnectionClass:
    """
    Data structure that contains connection
    """
    cdef clib.dxf_connection_t connection
    # sub_ptr_list contains pointers to all subscriptions related to current connection
    cdef cppdq[clib.dxf_subscription_t *] sub_ptr_list
    # each subscription has its own index in a list
    cdef int subs_order

    def __init__(self):
        self.subs_order = 0

    def __dealloc__(self):
        dxf_close_connection(self)

    cpdef SubscriptionClass make_new_subscription(self, data_len: int):
        cdef SubscriptionClass out = SubscriptionClass(data_len)
        out.connection = self.connection
        self.sub_ptr_list.push_back(&out.subscription)  # append pointer to new subscription
        out.subscription_order = self.subs_order  # assign each subscription an index
        self.subs_order += 1
        out.con_sub_list_ptr = &self.sub_ptr_list  # reverse pointer to pointers list
        return out


cdef class SubscriptionClass:
    """
    Data structure that contains subscription and related fields
    """
    cdef clib.dxf_connection_t connection
    cdef clib.dxf_subscription_t subscription
    cdef int subscription_order  # index in list of subscription pointers
    cdef cppdq[clib.dxf_subscription_t *] *con_sub_list_ptr  # pointer to list of subscription pointers
    cdef dxf_event_listener_t listener
    cdef object event_type_str
    cdef list columns
    cdef object data
    cdef void *u_data

    def __init__(self, data_len: int):
        """
        Parameters
        ----------
        data_len: int
            Sets maximum amount of events, that are kept in Subscription class
        """
        self.subscription = NULL
        self.columns = list()
        if data_len > 0:
            self.data = deque_wl(maxlen=data_len)
        else:
            self.data = deque_wl()
        self.u_data = <void *> self.data
        self.listener = NULL

    def __dealloc__(self):
        if self.subscription:  # if connection is not closed
            clib.dxf_close_subscription(self.subscription)
            self.con_sub_list_ptr[0][self.subscription_order] = NULL

    def get_data(self):
        """
        Method returns list with data, specified in event listener and returned data will be removed from object buffer

        Returns
        -------
        list
            List with data
        """
        return self.data.safe_get()

    def to_dataframe(self, keep: bool=True):
        """
        Method converts data to the Pandas DataFrame

        Parameters
        ----------
        keep: bool
            When True copies data to dataframe, otherwise pops. Default True

        Returns
        -------
        df: pandas DataFrame
        """
        if keep:
            df_data = self.data.copy()
        else:
            df_data = self.data.safe_get()

        df = pd.DataFrame(df_data, columns=self.columns)
        time_columns = df.columns[df.columns.str.contains('Time')]
        for column in time_columns:
            df.loc[:, column] = df.loc[:, column].astype('<M8[ms]')
        return df

def dxf_create_connection(address: Union[str, unicode, bytes] = 'demo.dxfeed.com:7300'):
    """
    Function creates connection to dxfeed given url address

    Parameters
    ----------
    address: str
        dxfeed url address

    Returns
    -------
    cc: ConnectionClass
        Cython ConnectionClass with information about connection
    """
    cc = ConnectionClass()
    address = address.encode('utf-8')
    clib.dxf_create_connection(address, NULL, NULL, NULL, NULL, NULL, &cc.connection)
    error_code = process_last_error(verbose=False)
    if error_code:
        raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
    return cc

def dxf_create_connection_auth_bearer(address: Union[str, unicode, bytes],
                                     token: Union[str, unicode, bytes]):
    """
    Function creates connection to dxfeed given url address and token

    Parameters
    ----------
    address: str
        dxfeed url address
    token: str
        dxfeed token

    Returns
    -------
    cc: ConnectionClass
        Cython ConnectionClass with information about connection
    """
    cc = ConnectionClass()
    address = address.encode('utf-8')
    token = token.encode('utf-8')
    clib.dxf_create_connection_auth_bearer(address, token,
                                          NULL, NULL, NULL, NULL, NULL, &cc.connection)
    error_code = process_last_error(verbose=False)
    if error_code:
        raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
    return cc

def dxf_create_subscription(ConnectionClass cc, event_type: str, candle_time: Optional[str] = None,
                            data_len: int = 100000):
    """
    Function creates subscription and writes all relevant information to SubscriptionClass

    Parameters
    ----------
    cc: ConnectionClass
        Variable with connection information
    event_type: str
        Event types: 'Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',
        'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration' or ''
    candle_time: str
        String of %Y-%m-%d %H:%M:%S datetime format for retrieving candles. By default set to now
    data_len: int
        Sets maximum amount of events, that are kept in Subscription class

    Returns
    -------
    sc: SubscriptionClass
        Cython SubscriptionClass with information about subscription
    """
    if not cc.connection:
        raise ValueError('Connection is not valid')
    if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',
                          'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration', ]:
        raise ValueError('Incorrect event type!')

    sc = cc.make_new_subscription(data_len=data_len)
    sc.event_type_str = event_type
    et_type_int = event_type_convert(event_type)

    try:
        candle_time = datetime.strptime(candle_time, '%Y-%m-%d %H:%M:%S') if candle_time else datetime.utcnow()
        timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000
    except ValueError:
        raise Exception("Inapropriate date format, should be %Y-%m-%d %H:%M:%S")

    if event_type == 'Candle':
        clib.dxf_create_subscription_timed(sc.connection, et_type_int, timestamp, &sc.subscription)
    else:
        clib.dxf_create_subscription(sc.connection, et_type_int, &sc.subscription)

    error_code = process_last_error(verbose=False)
    if error_code:
        raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
    return sc

def dxf_add_symbols(SubscriptionClass sc, symbols: Iterable[str]):
    """
    Adds symbols to subscription

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription
    symbols: list
        List of symbols to add
    """
    if not sc.subscription:
        raise ValueError('Subscription is not valid')
    for idx, sym in enumerate(symbols):
        if not isinstance(sym, str):
            warn(f'{sym} has type different from string')
            continue
        if not clib.dxf_add_symbol(sc.subscription, dxf_const_string_t_from_unicode(sym)):
            process_last_error()

def dxf_attach_listener(SubscriptionClass sc):
    """
    Function attaches default listener according to subscription type

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription
    """
    if not sc.subscription:
        raise ValueError('Subscription is not valid')
    if sc.event_type_str == 'Trade':
        sc.columns = lis.TRADE_COLUMNS
        sc.listener = lis.trade_default_listener
    elif sc.event_type_str == 'Quote':
        sc.columns = lis.QUOTE_COLUMNS
        sc.listener = lis.quote_default_listener
    elif sc.event_type_str == 'Summary':
        sc.columns = lis.SUMMARY_COLUMNS
        sc.listener = lis.summary_default_listener
    elif sc.event_type_str == 'Profile':
        sc.columns = lis.PROFILE_COLUMNS
        sc.listener = lis.profile_default_listener
    elif sc.event_type_str == 'TimeAndSale':
        sc.columns = lis.TIME_AND_SALE_COLUMNS
        sc.listener = lis.time_and_sale_default_listener
    elif sc.event_type_str == 'Candle':
        sc.columns = lis.CANDLE_COLUMNS
        sc.listener = lis.candle_default_listener
    elif sc.event_type_str == 'Order':
        sc.columns = lis.ORDER_COLUMNS
        sc.listener = lis.order_default_listener
    elif sc.event_type_str == 'TradeETH':
        sc.columns = lis.TRADE_COLUMNS
        sc.listener = lis.trade_default_listener
    elif sc.event_type_str == 'SpreadOrder':
        sc.columns = lis.ORDER_COLUMNS
        sc.listener = lis.order_default_listener
    elif sc.event_type_str == 'Greeks':
        sc.columns = lis.GREEKS_COLUMNS
        sc.listener = lis.greeks_default_listener
    elif sc.event_type_str == 'TheoPrice':
        sc.columns = lis.THEO_PRICE_COLUMNS
        sc.listener = lis.theo_price_default_listener
    elif sc.event_type_str == 'Underlying':
        sc.columns = lis.UNDERLYING_COLUMNS
        sc.listener = lis.underlying_default_listener
    elif sc.event_type_str == 'Series':
        sc.columns = lis.SERIES_COLUMNS
        sc.listener = lis.series_default_listener
    elif sc.event_type_str == 'Configuration':
        sc.columns = lis.CONFIGURATION_COLUMNS
        sc.listener = lis.configuration_default_listener
    else:
        raise Exception(f'No default listener for {sc.event_type_str} event type')

    if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):
        process_last_error()

def dxf_attach_custom_listener(SubscriptionClass sc, lis.FuncWrapper fw, columns: Iterable[str], data: Iterable = None):
    """
    Attaches custom listener

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription
    fw: FuncWrapper
        c function wrapped in FuncWrapper class with Cython
    columns: list
        Columns for internal data of SubscriptionClass
    data: dict
        Dict with new internal data structure of  SubscriptionClass
    """
    if not sc.subscription:
        raise ValueError('Subscription is not valid')
    if data:
        sc.data = data
    sc.columns = columns
    sc.listener = fw.func
    if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):
        process_last_error()

def dxf_detach_listener(SubscriptionClass sc):
    """
    Detaches any listener

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription
    """
    if not sc.subscription:
        raise ValueError('Subscription is not valid')
    if not clib.dxf_detach_event_listener(sc.subscription, sc.listener):
        process_last_error()

def dxf_close_connection(ConnectionClass cc):
    """
    Closes connection


    Parameters
    ----------
    cc: ConnectionClass
        Variable with connection information
    """

    # close all subscriptions before closing the connection
    for i in range(cc.sub_ptr_list.size()):
        if cc.sub_ptr_list[i]:  # subscription should not be closed previously
            clib.dxf_close_subscription(cc.sub_ptr_list[i][0])
            cc.sub_ptr_list[i][0] = NULL  # mark subscription as closed

    cc.sub_ptr_list.clear()

    if cc.connection:
        clib.dxf_close_connection(cc.connection)
        cc.connection = NULL

def dxf_close_subscription(SubscriptionClass sc):
    """
    Closes subscription

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription
    """
    if sc.subscription:
        clib.dxf_close_subscription(sc.subscription)
        sc.subscription = NULL

def dxf_get_current_connection_status(ConnectionClass cc, return_str: bool=True):
    """
    Returns one of four possible statuses

    Parameters
    ----------
    cc: ConnectionClass
        Variable with connection information
    return_str: bool
        When True returns connection status in string format, otherwise internal c representation as integer


    """
    status_mapping = {
        0: 'Not connected',
        1: 'Connected',
        2: 'Login required',
        3: 'Connected and authorized'
    }

    cdef clib.dxf_connection_status_t status
    clib.dxf_get_current_connection_status(cc.connection, &status)
    result = status
    if return_str:
        result = status_mapping[status]

    return result

def dxf_get_current_connected_address(ConnectionClass cc):
    """
    Returns current connected address

    Parameters
    ----------
    cc: ConnectionClass
        Variable with connection information

    Returns
    -------
    address: str
        Current connected address
    """
    if not cc.connection:
        raise ValueError('Connection is not valid')

    cdef char * address
    clib.dxf_get_current_connected_address(cc.connection, &address)
    return (<bytes>address).decode('UTF-8')

def dxf_initialize_logger(file_name: str, rewrite_file: bool, show_timezone_info: bool, verbose: bool):
    """
    Initializes the internal logger.
    Various actions and events, including the errors, are being logged throughout the library. They may be stored
    into the file.

    Parameters
    ----------
    file_name: str
        A full path to the file where the log is to be stored
    rewrite_file: bool
        A flag defining the file open mode if it's True then the log file will be rewritten
    show_timezone_info: bool
        A flag defining the time display option in the log file if it's True then the time will be displayed
        with the timezone suffix
    verbose: bool
        A flag defining the logging mode if it's True then the verbose logging will be enabled

    """
    clib.dxf_initialize_logger(file_name.encode('UTF-8'), int(rewrite_file), int(show_timezone_info), int(verbose))

def dxf_get_subscription_event_types(SubscriptionClass sc, return_str: bool=True):
    """
    Gets subscription event type

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription
    return_str: bool
        When True returns event type in string format, otherwise internal c representation as integer

    Returns
    -------
    str or int
        Subscription type
    """
    if not sc.subscription:
        raise ValueError('Invalid subscription')

    cdef int event_type

    et_mapping = {
        1: 'Trade',
        2: 'Quote',
        4: 'Summary',
        8: 'Profile',
        16: 'Order',
        32: 'TimeAndSale',
        64: 'Candle',
        128: 'TradeETH',
        256: 'SpreadOrder',
        512: 'Greeks',
        1024: 'TheoPrice',
        2048: 'Underlying',
        4096: 'Series',
        8192: 'Configuration',
        -16384: ''
    }

    clib.dxf_get_subscription_event_types (sc.subscription, &event_type)
    result = event_type
    if return_str:
        result = et_mapping[event_type]

    return result

def dxf_get_symbols(SubscriptionClass sc):
    """
    Retrieves the list of symbols currently added to the subscription.

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription

    Returns
    -------
    list
        List of unicode strings of subscription symbols
    """
    if not sc.subscription:
        raise ValueError('Invalid subscription')

    cdef dxf_const_string_t * symbols
    symbols_list = list()
    cdef int symbol_count
    cdef int i

    clib.dxf_get_symbols(sc.subscription, &symbols, &symbol_count)
    for i in range(symbol_count):
        symbols_list.append(unicode_from_dxf_const_string_t(symbols[i]))

    return symbols_list

def dxf_remove_symbols(SubscriptionClass sc, symbols: Iterable[str]):
    """
    Removes several symbols from the subscription

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription
    symbols: list
        List of symbols to remove
    """
    if not sc.subscription:
        raise ValueError('Invalid subscription')

    for symbol in symbols:
        clib.dxf_remove_symbol(sc.subscription, dxf_const_string_t_from_unicode(symbol))

def dxf_clear_symbols(SubscriptionClass sc):
    """
    Removes all symbols from the subscription

    Parameters
    ----------
    sc: SubscriptionClass
        SubscriptionClass with information about subscription
    """
    if not sc.subscription:
        raise ValueError('Invalid subscription')

    clib.dxf_clear_symbols(sc.subscription)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dxfeed',
 'dxfeed.core',
 'dxfeed.core.listeners',
 'dxfeed.core.pxd_include',
 'dxfeed.core.utils']

package_data = \
{'': ['*'],
 'dxfeed': ['dxfeed-c-api/*',
            'dxfeed-c-api/include/*',
            'dxfeed-c-api/include/pthreads/*',
            'dxfeed-c-api/src/*']}

install_requires = \
['pandas>=0.25.1,<0.26.0']

setup_kwargs = {
    'name': 'dxfeed',
    'version': '0.1.0',
    'description': 'DXFeed Python API via C API',
    'long_description': "# dxfeed package\n\nThis package provides access to [dxFeed](https://www.dxfeed.com/) streaming data.\nThe library is build as a thin wrapper over [dxFeed C-API library](https://github.com/dxFeed/dxfeed-c-api).\nWe use [Cython](https://cython.org/) in this project as it combines flexibility, reliability and\nusability in writing C extensions.\n\nThis package already contains basic C-API functions related to creating connections, subscriptions etc.\nMoreover default listeners (functions responsible for event processing) are ready to use. The user is also able to\nwrite his own custom listener in Cython\n\n## Installation\n\n**Requirements:** python >3.6, cython, pandas\n\n```python\npip3 install cython pandas\n```\n\nInstall package itself with PyPI\n\n```python\npip3 install dxfeed\n```\n\nor with tar.gz artifact\n\n```python\npip3 install dxfeed-x.x.x.tar.gz\n``` \n\n---\n**NOTE**\n\nMacOS is not supported yet\n\n---\n\n## Basic usage\n\nAll the functions in C API have similar ones in Python with the same name. Not all arguments are\nsupported by now, this work is in progress.\n\n**Import dxfeed library**:\n\n```python\nimport dxfeed as dx\n``` \n\n**Create connection**:\n\n```python\ncon = dx.dxf_create_connection(address='demo.dxfeed.com:7300')\n```\n\n**Create one or several subscriptions of certain event types**:\n```python\nsub1 = dx.dxf_create_subscription(con, 'Trade')\nsub2 = dx.dxf_create_subscription(con, 'Quote')\n```\n'Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH', 'SpreadOrder',\n'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration' event types are supported.\n\n**Add tickers you want to get data for**:\n```python\ndx.dxf_add_symbols(sub1, ['AAPL', 'MSFT'])\ndx.dxf_add_symbols(sub2, ['AAPL', 'C'])\n```\n\n**Attach listeners**:\n```python\ndx.dxf_attach_listener(sub1)\ndx.dxf_attach_listener(sub2)\n```\n\n`dxfeed` has default listeners for each event type, but you are able to write \nyour custom one. You can find how to do it at `example/Custom listener example.ipynb`.\n\n**Look at the data**:\n```python\nsub1.get_data()\nsub2.get_data()\n```\nThe data is stored in Subscription class. You can also turn dict to pandas DataFrame simply calling\n`sub1.to_dataframe()`.\n\n**Detach the listener, if you want to stop recieving data**:\n```python\ndx.dxf_detach_listener(sub1)\ndx.dxf_detach_listener(sub2)\n```\n\n**Finally, close your connection**:\n```python\ndx.dxf_close_connection(con)\n```\n",
    'author': 'Ivan Mishalkin',
    'author_email': 'ivan.mishalkin@devexperts.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



# from multiprocessing import Lock
from collections import deque
from typing import Any
from threading import Lock


class DequeWithLock(deque):
    """
    Class that provides lock mechanism to deque from collections for append, copy and get operations
    """
    def __init__(self, iterable=(), maxlen=None):
        super().__init__(iterable, maxlen)
        self.lock = Lock()

    def safe_append(self, data: Any):
        """
        Method appends data while locked

        Parameters
        ----------
        data: any
            Data to append
        """
        try:
            self.lock.acquire()
            self.append(data)
        finally:
            self.lock.release()

    def safe_get(self):
        """
        Method that pops all the data with subsequent clearing

        Returns
        -------
        list_to_return: list
            List filled with data
        """
        list_to_return = list()
        try:
            self.lock.acquire()
            list_to_return = self.copy()
            self.clear()
        finally:
            self.lock.release()
        return list(list_to_return)

